/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#ifndef __GNUC__
#define __attribute__(x)
#endif

#include <stdio.h>
#include <string.h>
#include "ServiceDiscoveryIfce.h"
#include "gfal_is_ifce_internal.h"
#include "Common.h"
#include "Definitions.h"

/*externs from FILE*/
extern SDService *sd_file_getService(const char *name, SDException *exception);
extern SDServiceDetails *sd_file_getServiceDetails(const char *name,
        SDException *exception);
extern SDServiceDataList *sd_file_getServiceData(const char *name,
        SDException *exception);
extern SDServiceList *sd_file_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);
extern SDServiceList *sd_file_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);
extern void sd_file_freeServiceDataList(SDServiceDataList *serviceDataList);
extern void sd_file_freeService(SDService *service);
extern void sd_file_freeServiceList(SDServiceList *serviceList);
extern void sd_file_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);
extern void sd_file_freeVOList(SDVOList *vos);
extern void sd_file_freeServiceDetails(SDServiceDetails *details);
extern void sd_file_sortbyVO(SDServiceDetailsList *servicedetailsList, const SDVOList *vos);

/*externs from BDII*/
extern SDServiceDetails *sd_bdii_getServiceDetails(const char *name,
        SDException *exception);
extern SDServiceList *sd_bdii_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);
extern SDServiceList *sd_bdii_listServicesByHost(const char *type, const char *host,
        const SDVOList *vos, SDException *exception);
extern SDServiceList *sd_bdii_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception);
extern SDServiceList *sd_bdii_listServices(const char *type, const char *site, const SDVOList *vos, SDException *exception);
extern SDService *sd_bdii_getService(const char *name, SDException *exception);
extern SDServiceDataList *sd_bdii_getServiceData(const char *name,
        SDException *exception);
extern char *sd_bdii_getServiceSite(const char *name, SDException *exception);
extern char *sd_bdii_getServiceWSDL(const char *name, SDException *exception);
extern char *sd_bdii_getServiceDataItem(const char *name, const char *key,
        SDException *exception);
extern void sd_bdii_freeService(SDService *service);
extern void sd_bdii_freeServiceList(SDServiceList *list);
extern void sd_bdii_freeServiceDataList(SDServiceDataList *list);
extern void sd_bdii_freeServiceDetails(SDServiceDetails *details);
extern void sd_bdii_freeVOList(SDVOList *vos);
extern void sd_bdii_freeServiceDetailsList(SDServiceDetailsList *servicedetailsList);



/*externs for gfal API*/
extern int gfal_sd_get_rls_endpoints(char **lrc_endpoint, char **rmc_endpoint, char *vo);
extern int gfal_sd_is_online(char *host, char *errbuf, int errbufsz);
extern int gfal_sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz);
extern int gfal_sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz);
extern int gfal_sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz);
extern int gfal_sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz);
extern int gfal_sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz);
extern void gfal_set_gfal_vo(char *vo);
extern void gfal_set_gfal_fqan(char **fqan, int fqan_size);

/*PUBLIC APIs*/
void SD_freeException(SDException *exception) {
    if (exception) {
        if (exception->reason &&
                exception->reason != SD_EMPTY_MESSAGE &&
                exception->reason != SD_NOMEM_MESSAGE) {
            free(exception->reason);
            exception->reason = NULL;
        }
        /* don't free exception itself, it's not on the heap */
    }
}

void SD_freeServiceDetails(SDServiceDetails *details) {
    if (!details)
        return;
    if(details->name)
    g_free(details->name);
    if(details->type)
    g_free(details->type);
    if(details->endpoint)
    g_free(details->endpoint);
    if(details->version)
    g_free(details->version);
    if(details->site)
    g_free(details->site);
    if(details->wsdl)
    g_free(details->wsdl);
    if(details->administration)
    g_free(details->administration);
    if(details->vos)
    SD_freeVOList(details->vos);
    if(details->associatedServices)
    SD_freeServiceList(details->associatedServices);
    if(details->data)
    SD_freeServiceDataList(details->data);
    g_free(details);
}

/**
 * Default implementation of the freeServiceDataList method.
 *
 * @param  A SD_freeServiceDataList.
 */

void SD_freeServiceDataList(SDServiceDataList *list) {
    int i;

    if (!list)
        return;
    for (i = 0; i < list->numItems; i++) {
        if (list->items[i].key) {
            g_free(list->items[i].key);
            list->items[i].key = NULL;
        }
        if (list->items[i].value) {
            g_free(list->items[i].value);
            list->items[i].value = NULL;
        }
    }

    g_free(list->items);
    list->items = NULL;
    
    g_free(list);
    list = NULL;
}

/**
 * Default implementation of the freeService method.
 *
 * @param  A SD_freeService.
 */

void SD_freeService(SDService *service) {
    if (!service)
        return;
	
    if(service->name){
	    g_free(service->name);
	    service->name = NULL;
    }
    if(service->type)    {
	    g_free(service->type);
	    service->type = NULL;
    }
    if(service->endpoint)   {     
	    g_free(service->endpoint);
	    service->endpoint = NULL;
    }
    if(service->version){        
	    g_free(service->version);
	    service->version = NULL;
    }
    
    if(service){
	    g_free(service);
	    service = NULL;
    }
}

/**
 * Default implementation of the freeServiceList method.
 *
 * @param  As SD_freeServiceList.
 */

void SD_freeServiceList(SDServiceList *list) {
    int i;

    if (!list)
        return;

    for (i = 0; i < list->numServices; i++){
       if(list->services && list->services[i])
        	sd_bdii_freeService(list->services[i]);
    }
    
    if(list->services){
	    g_free(list->services);
	    list->services = NULL;
    }	    

    if(list){
    	g_free(list);
    	list = NULL;
    }
}

char *SD_getServiceDataItem(const char *name, const char *key,
        SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    SDServiceDataList *list = NULL;
    char *value = NULL;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    if (NULL == plugin) /*env is not set*/ {
        list = sd_file_getServiceData(name, exception);
        if(!list){
		SD_freeException(exception);
		sd_file_freeServiceDataList(list);
		value = sd_bdii_getServiceDataItem(name, key, exception);
		if(!value){
			SD_freeException(exception);
			serviceDetails = SD_getServiceDetails(name, exception);
		}
	}

   
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');

        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims);
            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    list = sd_file_getServiceData(name, exception);
                    if (list) {
                        break;
                    }
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    value = sd_bdii_getServiceDataItem(name, key, exception);
                    if (value) {
                        break;
                    }
                } else {
                    SD_freeException(exception);
                    serviceDetails = SD_getServiceDetails(name, exception);
                    if (serviceDetails) {
                        break;
                    }
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (plugin && 0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                list = sd_file_getServiceData(name, exception);
            } else if (plugin && 0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                list = sd_bdii_getServiceData(name, exception);
            } else {
                SD_freeException(exception);
                serviceDetails = SD_getServiceDetails(name, exception);
            }
        }
    }


    if (serviceDetails)
        list = serviceDetails->data;
    if (list) {
        /* If we are emulating the getServiceDataItem method, we have to
         * look up the key and make a copy of the value */
        int i;

        value = NULL;
        for (i = 0; i < list->numItems; i++) {
            if (!strcmp(list->items[i].key, key)) {
                value = dupstr(list->items[i].value);
                if (!value)
                    SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
                break; 
            }
        }
    }

    if (serviceDetails)
        SD_freeServiceDetails(serviceDetails);
    else if (list)
        SD_freeServiceDataList(list);


    if (plugin)
        free(plugin);
    return value;
}

char *SD_getServiceSite(const char *name, SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    char *site = NULL;

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    site = sd_bdii_getServiceSite(name, exception);
    if (!site) {
        SD_freeException(exception);
        serviceDetails = SD_getServiceDetails(name, exception);
    }

    if (serviceDetails) {
        /* If we are emulating the getServiceSite method, we have to
         * make a copy of the value */
        if (serviceDetails->site) {
            site = dupstr(serviceDetails->site);
            if (!site)
                SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
        } else
            site = NULL;

        SD_freeServiceDetails(serviceDetails);
    }

    return site;
}

char *SD_getServiceWSDL(const char *name, SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    char *wsdl = NULL;

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);


    wsdl = sd_bdii_getServiceWSDL(name, exception);
    if (!wsdl) {
        SD_freeException(exception);
        serviceDetails = SD_getServiceDetails(name, exception);
    }


    if (serviceDetails) {
        /* If we are emulating the getServiceWSDL method, we have to
         * make a copy of the value */
        if (serviceDetails->wsdl) {
            wsdl = dupstr(serviceDetails->wsdl);
            if (!wsdl)
                SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
        } else
            wsdl = NULL;

        SD_freeServiceDetails(serviceDetails);
    }

    return wsdl;
}

SDServiceDataList *SD_getServiceData(const char *name,
        SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    SDServiceDataList *list = NULL;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);


    if (NULL == plugin) /*env is not set, query file,bdii with this order */ {
        list = sd_file_getServiceData(name, exception);
        if (!list) {
	    sd_file_freeServiceDataList(list);	     
            SD_freeException(exception);
            list = sd_bdii_getServiceData(name, exception);
            if (!list) {
	        sd_bdii_freeServiceDataList(list);	     
                SD_freeException(exception);
                serviceDetails = SD_getServiceDetails(name, exception);
            }
        }
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');

        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims);
            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    list = sd_file_getServiceData(name, exception);
                    if (list) {
                        break;
                    }
		    else
		    	sd_file_freeServiceDataList(list);
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    list = sd_bdii_getServiceData(name, exception);
                    if (list) {
                        break;
                    }
		    else
		    	sd_bdii_freeServiceDataList(list);
                } else {
                    SD_freeException(exception);
                    serviceDetails = SD_getServiceDetails(name, exception);
                    if (serviceDetails) {
                        break;
                    }
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (plugin && 0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                list = sd_file_getServiceData(name, exception);
            } else if (plugin && 0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                list = sd_bdii_getServiceData(name, exception);
            } else {
                serviceDetails = SD_getServiceDetails(name, exception);
            }
        }
    }




    if (serviceDetails) {
        /* If we are emulating the getServiceData method, we have to make
         * a copy of the data list */
        int i;

        list = (SDServiceDataList *) malloc(sizeof (SDServiceDataList));
        if (!list)
            SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
        else {
            list->numItems = serviceDetails->data->numItems;
            list->items = (SDServiceData *) calloc(list->numItems, sizeof (SDServiceData));
            if (!list->items) {
                SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
                SD_freeServiceDataList(list);
                list = NULL;
            }
        }

        for (i = 0; list && i < list->numItems; i++) {
            list->items[i].key = dupstr(serviceDetails->data->items[i].key);
            list->items[i].value = dupstr(serviceDetails->data->items[i].value);
            if (!list->items[i].key || !list->items[i].value) {
                SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
                SD_freeServiceDataList(list);
                list = NULL;
                break;
            }
        }

        SD_freeServiceDetails(serviceDetails);
    }
    if (plugin)
        free(plugin);
    return list;
}

SDService *SD_getService(const char *name, SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    SDService *service = NULL;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    if (NULL == plugin) /*env is not set, query file,bdii with this order */ {
        service = sd_file_getService(name, exception);
        if (!service) {
	    sd_file_freeService(service);
            SD_freeException(exception);
            service = sd_bdii_getService(name, exception);
            if (!service) {
	    	sd_bdii_freeService(service);
                SD_freeException(exception);
                serviceDetails = SD_getServiceDetails(name, exception);
		if(!serviceDetails){		        
			SD_freeServiceDetails(serviceDetails);
		}
            }
        }
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');
        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims);
            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    service = sd_file_getService(name, exception);
                    if (service) {
                        break;
                    }
		    else
		    	    sd_file_freeService(service);
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    service = sd_bdii_getService(name, exception);
                    if (service) {
                        break;
                    }
		    else
		    	sd_bdii_freeService(service);
                } else {
                    SD_freeException(exception);
                    serviceDetails = SD_getServiceDetails(name, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
			SD_freeServiceDetails(serviceDetails);
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (plugin && 0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                service = sd_file_getService(name, exception);
                if (!service) 
	            sd_file_freeService(service);
            } else if (plugin && 0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                service = sd_bdii_getService(name, exception);
		if(!service)
			sd_bdii_freeService(service);
            } else {
                SD_freeException(exception);
                serviceDetails = SD_getServiceDetails(name, exception);
            }
        }
    }

    if (serviceDetails) {
        /* If we are emulating the getService method, we have to make
         * a copy of the data */
        service = (SDService *) calloc(1, sizeof (SDService));
        if (!service)
            SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
        else {
            /* Copy the basic service data */
            service->name = dupstr(serviceDetails->name);
            service->type = dupstr(serviceDetails->type);
            service->endpoint = dupstr(serviceDetails->endpoint);
            service->version = dupstr(serviceDetails->version);

            /* Check for failed dupstr()s */
            if (!service->name || !service->type || !service->endpoint ||
                    !service->version) {
                SD_setException(exception, SDStatus_FAILURE, ERROR_OUTOFMEMORY, NULL, 0);
                SD_freeService(service);
                service = NULL;
            }
        }

        SD_freeServiceDetails(serviceDetails);
    }

    if (plugin)
        free(plugin);

    return service;
}

SDServiceList *SD_listServices(const char *type, const char *site, const SDVOList *vos, SDException *exception) {
    SDServiceList *result = NULL;
    SDVOList *tmpvos;

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    if (!site) {
        site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0))
            site = NULL;
    }
    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();


    result = SD_listServicesByData(NULL, type, site,
            tmpvos ? tmpvos : vos,
            exception);

    if (tmpvos)
        SD_freeVOList(tmpvos);

    return result;
}

SDServiceList *SD_listServicesByData(const SDServiceDataList *data,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    SDServiceList *serviceDetails = NULL;
    SDVOList *tmpvos;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);


    if (!site) {
        site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0))
            site = NULL;
    }

    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();


    if (NULL == plugin) /*env is not set, query file,bdii with this order */ {
        serviceDetails = sd_file_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
        if (!serviceDetails) {
	    sd_file_freeServiceList(serviceDetails);
            SD_freeException(exception);
            serviceDetails = sd_bdii_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
        }
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');

        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims); /*extract first token*/

            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_file_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
		    	    sd_file_freeServiceList(serviceDetails);
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_bdii_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
		    	    sd_bdii_freeServiceList(serviceDetails);
                } else {
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                serviceDetails = sd_file_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
		if(!serviceDetails)
			sd_file_freeServiceList(serviceDetails);
            } else if (0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                serviceDetails = sd_bdii_listServicesByData(data, type, site, tmpvos ? tmpvos : vos, exception);
		if(!serviceDetails)
			sd_bdii_freeServiceList(serviceDetails);
            }
        }
    }


    if (tmpvos)
        SD_freeVOList(tmpvos);
    if (plugin)
        free(plugin);

    return serviceDetails;
}





typedef int (*SD_matchEndpointHost)(const char *endpoint, const char *host);

/* matching '^host$' */

static int SD_matchEndpointHost_0(const char *endpoint, const char *host) {
    return !strcasecmp(endpoint, host);
}

/* matching '^host:' */

static int SD_matchEndpointHost_1(const char *endpoint, const char *host) {
    return (!strncasecmp(endpoint, host, strlen(host)))
            && (strlen(endpoint) > strlen(host))
            && (endpoint[strlen(host)] == ':');
}

/* matching '://host/' */

static int SD_matchEndpointHost_2(const char *endpoint, const char *host) {
    const char *hostp = strcasestr(endpoint, host);

    if (NULL == hostp) return 0; /* host not found */

    int index = hostp - endpoint;
    if (index < 3) return 0; /*  cannot match '://' */
    if (strlen(endpoint) < index + strlen(host) + 1) return 0; /* last slash */
    return (hostp[-3] == ':') && (hostp[-2] == '/') && (hostp[-1] == '/')
            && (hostp[strlen(host)] == '/');
}

static int SD_matchEndpointHost_3(const char *endpoint, const char *host) {
    const char *hostp = strcasestr(endpoint, host);

    if (NULL == hostp) return 0; /* host not found */

    int index = hostp - endpoint;
    if (index < 3) return 0; /*  cannot match '://' */
    if (strlen(endpoint) < index + strlen(host) + 1) return 0; 
    return (hostp[-3] == ':') && (hostp[-2] == '/') && (hostp[-1] == '/')
            && (hostp[strlen(host)] == ':');
}

static int SD_matchEndpointHost_4(const char *endpoint, const char *host) {
    return strcasestr(endpoint, host) != NULL;
}

static SD_matchEndpointHost SD_matchEndpointHostFunctions[] = {
    &SD_matchEndpointHost_0,
    &SD_matchEndpointHost_1,
    &SD_matchEndpointHost_2,
    &SD_matchEndpointHost_3,
    &SD_matchEndpointHost_4
};
static int SD_num_matches = sizeof (SD_matchEndpointHostFunctions) / sizeof (SD_matchEndpointHost);

SDServiceList *SD_listServicesByHost(const char *type, const char *host,
        const SDVOList *vos, SDException *exception) {

    SDServiceList *result = NULL;
    SDVOList *tmpvos = NULL;
    int i, j;
    int numServices;

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();


    result = sd_bdii_listServicesByHost(type, host, tmpvos ? tmpvos : vos, exception);

    if (!result) {
    
        sd_bdii_freeServiceList(result);
        SD_freeException(exception);
	
        /* filter the host and port from the result of */

        /* listServices on the client side */
        result = SD_listServices(type, NULL, tmpvos ? tmpvos : vos, exception);
        if (result != NULL && exception->status == SDStatus_SUCCESS) {
            /* Check for the host name in the endpoint in this order:
            //    1. ^host$
            //    2. ^host:
            //    3. /host:
            //    4. /host/
            //    5. host
            // Run the filtering with the first matching pattern.
            // TODO: implement the filtering in bdii-c */
            int f, fmatched;
            SD_matchEndpointHost matchFunction = NULL;
            for (f = 0; f < SD_num_matches; f++) {
                matchFunction = SD_matchEndpointHostFunctions[f];
                fmatched = 0;
                for (i = 0; i < result->numServices; i++) {
                    if (result->services[i]->endpoint != NULL) {
                        if (matchFunction(result->services[i]->endpoint, host)) {
                            fmatched++;
                        }
                    }
                } /* end for loop 
                // break the loop, if there was a match and use that for filtering */
                if (fmatched > 0) break;
            }
            /* in the worst case, when there was no match,
            // matchFunction will point to the last function */

            for (i = 0; i < result->numServices; i++) {
                if (result->services[i]->endpoint != NULL) {
                    /* check if the endpoint contains the host name */
                    if (!matchFunction(result->services[i]->endpoint, host)) {
                        /* no match. Delete the service entry and mark it as null 
                        // in the list */
                        SD_freeService(result->services[i]);
                        result->services[i] = NULL;
                    }
                }
            } /* end for loop 

            // Reorder the list so the nulls are at the end
            // Please Note: this algorithm may not be the best one but the 
            // number of expected elements should be quite limited, so it 
            // shouldn't be a real problem */
            numServices = 0;
            for (i = 0; i < result->numServices; ++i) {
                if (NULL == result->services[i]) {
                    /* Get the next valid one */
                    for (j = i + 1; j < result->numServices; ++j) {
                        if (NULL != result->services[j]) {
                            /* Swicth the two elements */
                            result->services[i] = result->services[j];
                            result->services[j] = 0;
                            ++numServices;
                            break;
                        }
                    }
                } else {
                    ++numServices;
                }
            } /* end reordering loop */
            result->numServices = numServices;
        } /* if result != NULL */
    }


    if (tmpvos)
        SD_freeVOList(tmpvos);

    return result;
}

SDServiceList *SD_listAssociatedServices(const char *name,
        const char *type, const char *site, const SDVOList *vos,
        SDException *exception) {
    SDServiceList *serviceDetails = NULL;
    SDVOList *tmpvos;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    if (!site) {
        site = getenv(SDEnv_SITE);
        /* treat empty string as if it does not exist */
        if (site && (strlen(site) == 0))
            site = NULL;
    }

    tmpvos = NULL;
    /* Let the user override the VO if the caller did not specify
     * one explicitely */
    if (!vos || !vos->numNames)
        tmpvos = checkVOEnv();

    if (NULL == plugin) /*env is not set, query file,bdii with this order */ {
        serviceDetails = sd_file_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);
        if (!serviceDetails) { 
	    sd_file_freeServiceList(serviceDetails);          
            SD_freeException(exception);
            serviceDetails = sd_bdii_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);            
        }
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');

        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims);
            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_file_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
		    	sd_file_freeServiceList(serviceDetails);
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_bdii_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
			    sd_file_freeServiceList(serviceDetails);
                } else {
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (plugin && 0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                serviceDetails = sd_file_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);
            } else if (plugin && 0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                serviceDetails = sd_bdii_listAssociatedServices(name, type, site, tmpvos ? tmpvos : vos, exception);
            }
        }
    }

    if (tmpvos)
        SD_freeVOList(tmpvos);
    if (plugin)
        free(plugin);

    return serviceDetails;
}

SDServiceDetails *SD_getServiceDetails(const char *name,
        SDException *exception) {
    SDServiceDetails *serviceDetails = NULL;
    char delims[] = ",";
    char *plugin = NULL;
    plugin = plugin_is_set();

    SD_setException(exception, SDStatus_SUCCESS, NULL, NULL, 0);

    if (plugin == NULL) /*env is not set, query file,bdii with this order */ {
        serviceDetails = sd_file_getServiceDetails(name, exception);
        if (!serviceDetails) {
    	    sd_file_freeServiceDetails(serviceDetails);
            SD_freeException(exception);
            serviceDetails = sd_bdii_getServiceDetails(name, exception);
        }
    } else /*env is set, check the order*/ {
        char *comma = strchr(plugin, ',');

        if (plugin && comma) {
            char *result = NULL;
            result = strtok(plugin, delims);
            while (result != NULL) {
                if (0 == strcmp(result, available_plungins[0])) /*file*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_file_getServiceDetails(name, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
		       sd_file_freeServiceDetails(serviceDetails);
                } else if (0 == strcmp(result, available_plungins[1])) /*bdii*/ {
                    SD_freeException(exception);
                    serviceDetails = sd_bdii_getServiceDetails(name, exception);
                    if (serviceDetails) {
                        break;
                    }
		    else
		       sd_bdii_freeServiceDetails(serviceDetails);
                } else {
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
            if (plugin && 0 == strcmp(plugin, available_plungins[0])) /*file*/ {
                SD_freeException(exception);
                serviceDetails = sd_file_getServiceDetails(name, exception);
            } else if (plugin && 0 == strcmp(plugin, available_plungins[1])) /*bdii*/ {
                SD_freeException(exception);
                serviceDetails = sd_bdii_getServiceDetails(name, exception);
            }
        }
    }

    if (plugin)
        free(plugin);

    return serviceDetails;
}

/*GFAL API STARTS HERE*/
int sd_get_rls_endpoints(char **lrc_endpoint, char **rmc_endpoint, char *vo) {
    return gfal_sd_get_rls_endpoints(lrc_endpoint, rmc_endpoint, vo);
}

int sd_is_online(char *host, char *errbuf, int errbufsz) {
    return gfal_sd_is_online(host, errbuf, errbufsz);
}

int sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz) {
    return gfal_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
}

int sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz) {
    return gfal_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
}

int sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz) {
    return gfal_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
}

int sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz) {
    return gfal_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
}

int sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz) {
    return gfal_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
}

void set_gfal_vo(char *vo) {
    gfal_set_gfal_vo(vo);
}

void set_gfal_fqan(char **fqan, int fqan_size) {
    gfal_set_gfal_fqan(fqan, fqan_size);
}
