/*                                                                -*- mode: c; c-basic-offset: 8;  -*-
 * Copyright (c) Members of the EGEE Collaboration. 2004.
 * See http://www.eu-egee.org/partners/ for details on the copyright holders.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*
 ============================================================================
 This is the inrernal interface to be used by gfal external API, in order 
 to differentiate between Glue1.3 and Glue2
 ============================================================================
 */


#include <stdio.h>
#include "ServiceDiscoveryIfce.h"
#include "gfal_is_ifce_internal.h"
#include "Common.h"
#include "Definitions.h"


/*externs for glue1.3*/
extern int g1_sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz);
extern int g1_sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz);
extern int g1_sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz);
extern int g1_sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz);
extern int g1_sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz);
extern int g1_sd_is_online(char *host, char *errbuf, int errbufsz);

/*externs for glue2*/
extern int g2_sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz);
extern int g2_sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz);
extern int g2_sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz);
extern int g2_sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz);
extern int g2_sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz);
extern int g2_sd_is_online(char *host, char *errbuf, int errbufsz);



/*Internal API*/
char *vo_global = NULL;
char **fqan_global = NULL;
int fqan_size_global = 0;

void gfal_set_gfal_vo(char *vo) {
    setenv("GLITE_SD_PLUGIN", "bdii", 1);
    vo_global = vo;
}

void gfal_set_gfal_fqan(char **fqan, int fqan_size) {
    setenv("GLITE_SD_PLUGIN", "bdii", 1);
    fqan_global = fqan;
    fqan_size_global = fqan_size;
}

int gfal_sd_get_ce_ap(const char *host, char **ce_ap, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else {
   		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;	
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_get_ce_ap(host, ce_ap, errbuf, errbufsz);
            else{
	    	SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);	    
                ret = -1;
		}
        }
    }

    if (base)
        free(base);

    return ret;
}

/* get from the BDII the endpoint for the LFC */

int gfal_sd_get_lfc_endpoint(char **lfc_endpoint, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else {
		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;	
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_get_lfc_endpoint(lfc_endpoint, errbuf, errbufsz);
            else{
	    	SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);	    
                ret = -1;
		}
        }
    }

    if (base)
        free(base);

    return ret;
}


int gfal_sd_get_storage_path(const char *host, const char *spacetokendesc,
        char **sa_path, char **sa_root, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else {
		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;	
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_get_storage_path(host, spacetokendesc, sa_path, sa_root, errbuf, errbufsz);
            else{
	    	SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                ret = -1;
		}
        }
    }

    if (base)
        free(base);

    return ret;

}


/**
 *  get from the BDII the SE type ( disk, srm_v1 )
 *
 *  */

int gfal_sd_get_se_types_and_endpoints(const char *host, char ***se_types,
        char ***se_endpoints, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
                    if (ret == 0)
                        break;                    
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
                    if (ret == 0)
                        break;                    
                } else {
		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_get_se_types_and_endpoints(host, se_types, se_endpoints, errbuf, errbufsz);
            else{
	    	SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                ret = -1;
		}
        }
    }

    if (base)
        free(base);

    return ret;
}


/* get from the BDII the list of supported protocols with their associated
 * port number
 */
int gfal_sd_get_seap_info(const char *host, char ***access_protocol, int **port, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else {
		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;	
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_get_seap_info(host, access_protocol, port, errbuf, errbufsz);
            else{
	    	SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                ret = -1;
		}
        }
    }

    if (base)
        free(base);

    return ret;
}

/*
This is required by Savannah task  31706: lcg-utils/GFAL should check GlueSEStatus.
If it is "Production" or missing then the API function returns true, else false. 
The parameter passed is the FQN of the SE
 */

int gfal_sd_is_online(char *host, char *errbuf, int errbufsz) {
    int ret = 0;
    char *comma = NULL;
    char *result = NULL;
    char delims[] = ",";
    SDException exception;
    char *base = base_is_set();

    if (base == NULL) {
        ret = g1_sd_is_online(host, errbuf, errbufsz);
    } else {
        comma = strchr(base, ',');
        if (base && comma) {
            result = strtok(base, delims);
            while (result != NULL) {
                if (0 == strcmp(result, "o=grid")) {
                    ret = g1_sd_is_online(host, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else if (0 == strcmp(result, "o=glue")) {
                    ret = g2_sd_is_online(host, errbuf, errbufsz);
                    if (ret == 0)
                        break;
                } else {
		    SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                    ret = -1;
                    break;
                }
                result = strtok(NULL, delims);
            }
        } else {
	    result = base;	
            if (0 == strcmp(result, "o=grid"))
                ret = g1_sd_is_online(host, errbuf, errbufsz);
            else if (0 == strcmp(result, "o=glue"))
                ret = g2_sd_is_online(host, errbuf, errbufsz);
            else{
	        SD_setException(&exception, SDStatus_FAILURE, "[BDII]: Check LCG_GFAL_BDII_DEFAULT_SEARCH_BASE env variable", errbuf, errbufsz);
                ret = -1;
	    }
        }
    }

    if (base)
        free(base);

    return ret;
}

/*RLS is not used anymore, hardcoded return value is used then*/
int gfal_sd_get_rls_endpoints(char **lrc_endpoint, char **rmc_endpoint, char *vo) {
    return (-1);
}
